/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "entity.h"
#include "assert/assert.h"
#include "util/memory.h"
#include "world/world.h"
#include "world/tiledata.h"
#include "physics/physics.h"

entity_t ENTITIES[ENTITY_COUNT_MAX] = {0};

entitycallback_t ENTITY_CALLBACKS[ENTITY_TYPE_COUNT] = {
  {NULL}, // ENTITY_TYPE_NULL
  {playerNPCInit, playerNPCUpdate}, // ENTITY_TYPE_PLAYER
  {npcInit, npcUpdate} // ENTITY_TYPE_NPC
};

void entityInit(entity_t *entity, const entitytype_t type) {
  assertNotNull(entity, "Entity pointer cannot be NULL");
  assertTrue(type != ENTITY_TYPE_NULL, "Entity type NULL");
  assertTrue(type < ENTITY_TYPE_COUNT, "Entity type out of bounds");
  assertNotNull(
    ENTITY_CALLBACKS[type].init,
    "Entity type has no init callback"
  );
  
  memoryZero(entity, sizeof(entity_t));

  entity->type = type;
  ENTITY_CALLBACKS[type].init(entity);
}

void entityUpdate(entity_t *entity) {
  assertNotNull(entity, "Entity pointer cannot be NULL");
  assertTrue(entity->type != ENTITY_TYPE_NULL, "Entity type NULL");
  assertTrue(entity->type < ENTITY_TYPE_COUNT, "Entity type out of bounds");
  assertNotNull(
    ENTITY_CALLBACKS[entity->type].update,
    "Entity type has no update callback"
  );

  ENTITY_CALLBACKS[entity->type].update(entity);

  if(
    entity->vx == FIXED248(0, 0) &&
    entity->vy == FIXED248(0, 0)
  ) return;

  fixed248_t newX = entity->x + entity->vx;
  fixed248_t newY = entity->y + entity->vy;
  fixed248_t halfTileWH = FIXED248(TILE_WIDTH_HEIGHT / 2, 0);
  fixed248_t selfCircR = halfTileWH;

  // Check for collisions with tiles
  uint8_t tileChecks = ENTITY_TILE_COLISSION_RESOLUTION_COUNT;
  do {
    // loop to resolve tile collisions multiple times to handle sliding
    // correctly
    bool tileResolved = false;

    // Compute affected tile range
    fixed248_t tileStartX = fx248Floor(fx248Divfx248(
      (newX - halfTileWH), FIXED248(TILE_WIDTH_HEIGHT, 0)
    ));
    fixed248_t tileStartY = fx248Floor(fx248Divfx248(
      (newY - halfTileWH), FIXED248(TILE_WIDTH_HEIGHT, 0)
    ));
    fixed248_t tileEndX = fx248Ceil(fx248Divfx248(
      (newX + halfTileWH), FIXED248(TILE_WIDTH_HEIGHT, 0)
    ));
    fixed248_t tileEndY = fx248Ceil(fx248Divfx248(
      (newY + halfTileWH), FIXED248(TILE_WIDTH_HEIGHT, 0)
    ));

    // For each tile
    for(fixed248_t y = tileStartY; y <= tileEndY; y += FIXED248_ONE) {
      for(fixed248_t x = tileStartX; x <= tileEndX; x += FIXED248_ONE) {
        uint16_t tileX = fx248Tou16(x);
        uint16_t tileY = fx248Tou16(y);
        uint16_t chunkX = tileX / CHUNK_WIDTH;
        uint16_t chunkY = tileY / CHUNK_HEIGHT;
        chunk_t *chunk = chunkGetChunkAt(chunkX, chunkY);
        if(chunk == NULL) continue;

        uint8_t chunkTileX = tileX % CHUNK_WIDTH;
        uint8_t chunkTileY = tileY % CHUNK_HEIGHT;
        tile_t tile = chunk->tilesBase[chunkTileY * CHUNK_WIDTH + chunkTileX];

        fixed248_t lx = fx248Mulfx248(x, FIXED248(TILE_WIDTH_HEIGHT, 0));
        fixed248_t ty = fx248Mulfx248(y, FIXED248(TILE_WIDTH_HEIGHT, 0));
        fixed248_t rx = fx248Addfx248(lx, FIXED248(TILE_WIDTH_HEIGHT, 0));
        fixed248_t by = fx248Addfx248(ty, FIXED248(TILE_WIDTH_HEIGHT, 0));
        
        // Determine tile collision type
        collisionresult_t collision;
        switch(TILE_META_DATA[tile].solidType) {
          case TILE_SOLID_FULL:
            collision = physicsCheckCircleAABB(
              fx248Addfx248(newX, halfTileWH), fx248Addfx248(newY, halfTileWH), selfCircR,
              lx, ty,
              FIXED248(TILE_WIDTH_HEIGHT, 0),
              FIXED248(TILE_WIDTH_HEIGHT, 0)
            );
            break;

          case TILE_SOLID_TRIANGLE_TOP_RIGHT:
            collision = physicsCheckCircleTriangle(
              fx248Addfx248(newX, halfTileWH), fx248Addfx248(newY, halfTileWH), selfCircR,
              rx, by,
              rx, ty,
              lx, ty
            );
            break;

          case TILE_SOLID_TRIANGLE_TOP_LEFT:
            collision = physicsCheckCircleTriangle(
              fx248Addfx248(newX, halfTileWH), fx248Addfx248(newY, halfTileWH), selfCircR,
              lx, by,
              lx, ty,
              rx, ty
            );
            break;

          case TILE_SOLID_TRIANGLE_BOTTOM_RIGHT:
            collision = physicsCheckCircleTriangle(
              fx248Addfx248(newX, halfTileWH), fx248Addfx248(newY, halfTileWH), selfCircR,
              rx, ty,
              rx, by,
              lx, by
            );
            break;
          case TILE_SOLID_TRIANGLE_BOTTOM_LEFT:
            collision = physicsCheckCircleTriangle(
              fx248Addfx248(newX, halfTileWH), fx248Addfx248(newY, halfTileWH), selfCircR,
              lx, ty,
              lx, by,
              rx, by
            );
            break;

          default:
            continue;
        }


        if(collision.hit && collision.depth > FIXED248(0, 1)) {
          fixed248_t slideX = fx248Mulfx248(
            collision.normalX, collision.depth
          );
          fixed248_t slideY = fx248Mulfx248(
            collision.normalY, collision.depth
          );
          newX -= slideX;
          newY -= slideY;
          tileResolved = true;
        }
      }
    }
    if(!tileResolved) break; // no more overlaps
  } while(--tileChecks > 0);

  // Check for collisions with other entities
  entity_t *otherEntity = ENTITIES;
  do {
    // Skip self and null entities
    if(otherEntity == entity || otherEntity->type == ENTITY_TYPE_NULL) continue;

    fixed248_t otherCircR = halfTileWH;

    collisionresult_t collision = physicsCheckCircleCircle(
      newX, newY, selfCircR,
      otherEntity->x, otherEntity->y, otherCircR
    );
    if(!collision.hit) continue;

    // Collision with entity detected. Slide out of collision.
    fixed248_t slideX = fx248Mulfx248(collision.normalX, collision.depth);
    fixed248_t slideY = fx248Mulfx248(collision.normalY, collision.depth);
    newX -= slideX;
    newY -= slideY;
  } while(++otherEntity < ENTITIES + ENTITY_COUNT_MAX);

  entity->x = newX;
  entity->y = newY;
}