/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "physics.h"

collisionresult_t physicsCheckCircleCircle(
  fixed248_t circle0x, fixed248_t circle0y, fixed248_t circle0r,
  fixed248_t circle1x, fixed248_t circle1y, fixed248_t circle1r
) {
  collisionresult_t result;
  // Compute vector between centers
  fixed248_t dx = fx248Subfx248(circle1x, circle0x);
  fixed248_t dy = fx248Subfx248(circle1y, circle0y);

  // Distance squared between centers
  fixed248_t distSq = fx248Addfx248(
    fx248Mulfx248(dx, dx),
    fx248Mulfx248(dy, dy)
  );

  // Sum of radii
  fixed248_t rSum = fx248Addfx248(circle0r, circle1r);
  fixed248_t rSumSq = fx248Mulfx248(rSum, rSum);

  if(distSq > rSumSq) {
    // No collision
    result.hit = false;
    return result;
  }

  // Collision: calculate normal and penetration depth
  fixed248_t dist = fx248Sqrt(distSq);

  // If centers are the same, pick arbitrary normal (1,0)
  if(dist == 0) {
    result.normalX = FIXED248(1,0);
    result.normalY = 0;
    result.depth = rSum;
  } else {
    // Normalized direction from circle0 to circle1
    result.normalX = fx248Divfx248(dx, dist);
    result.normalY = fx248Divfx248(dy, dist);
    // Penetration depth = sum of radii - distance
    result.depth = fx248Subfx248(rSum, dist);
  }
  result.hit = true;
  return result;
}

collisionresult_t physicsCheckCircleAABB(
  fixed248_t circleX, fixed248_t circleY, fixed248_t circleR,
  fixed248_t aabbX, fixed248_t aabbY,
  fixed248_t aabbWidth, fixed248_t aabbHeight
) {
  collisionresult_t result;

  // Find the closest point on the AABB to the circle center
  fixed248_t closestX = fx248Max(
    aabbX, fx248Min(circleX, fx248Addfx248(aabbX, aabbWidth))
  );
  fixed248_t closestY = fx248Max(
    aabbY, fx248Min(circleY, fx248Addfx248(aabbY, aabbHeight))
  );

  // Vector from circle center to closest point
  fixed248_t dx = fx248Subfx248(closestX, circleX);
  fixed248_t dy = fx248Subfx248(closestY, circleY);

  // Distance squared from circle center to closest point
  fixed248_t distSq = fx248Addfx248(fx248Mulfx248(dx, dx), fx248Mulfx248(dy, dy));

  // Check if distance is less than radius squared
  if(distSq > fx248Mulfx248(circleR, circleR)) {
    result.hit = false;
    return result;
  }

  // Collision: calculate normal and penetration depth
  fixed248_t dist = fx248Sqrt(distSq);
  
  if(dist <= FIXED248(0, 1)) {
    // Circle center is at the AABB corner
    result.normalX = FIXED248(1,0);
    result.normalY = FIXED248(0,1);
    result.depth = circleR;
  } else {
    // Normalized direction from circle center to closest point
    result.normalX = fx248Divfx248(dx, dist);
    result.normalY = fx248Divfx248(dy, dist);
    // Penetration depth = radius - distance
    result.depth = fx248Subfx248(circleR, dist);
  }
  
  result.hit = true;
  return result;
}

void physicsClosestPointOnSegment(
  fixed248_t ax, fixed248_t ay,
  fixed248_t bx, fixed248_t by,
  fixed248_t px, fixed248_t py,
  fixed248_t *outX, fixed248_t *outY
) {
  fixed248_t abx = fx248Subfx248(bx, ax);
  fixed248_t aby = fx248Subfx248(by, ay);
  fixed248_t apx = fx248Subfx248(px, ax);
  fixed248_t apy = fx248Subfx248(py, ay);

  fixed248_t abLenSq = fx248Addfx248(
    fx248Mulfx248(abx, abx),
    fx248Mulfx248(aby, aby)
  );

  if(abLenSq == FIXED248_ZERO) {
    *outX = ax;
    *outY = ay;
    return;
  }

  fixed248_t t = fx248Divfx248(
    fx248Addfx248(fx248Mulfx248(apx, abx), fx248Mulfx248(apy, aby)),
    abLenSq
  );

  if(t < FIXED248_ZERO) t = FIXED248_ZERO;
  if(t > FIXED248_ONE)  t = FIXED248_ONE;

  *outX = fx248Addfx248(ax, fx248Mulfx248(abx, t));
  *outY = fx248Addfx248(ay, fx248Mulfx248(aby, t));
}

bool_t physicsIsPointInTriangle(
  fixed248_t px, fixed248_t py,
  fixed248_t ax, fixed248_t ay,
  fixed248_t bx, fixed248_t by,
  fixed248_t cx, fixed248_t cy
) {
  fixed248_t abx = fx248Subfx248(bx, ax);
  fixed248_t aby = fx248Subfx248(by, ay);
  fixed248_t bcx = fx248Subfx248(cx, bx);
  fixed248_t bcy = fx248Subfx248(cy, by);
  fixed248_t cax = fx248Subfx248(ax, cx);
  fixed248_t cay = fx248Subfx248(ay, cy);

  fixed248_t apx = fx248Subfx248(px, ax);
  fixed248_t apy = fx248Subfx248(py, ay);
  fixed248_t bpx = fx248Subfx248(px, bx);
  fixed248_t bpy = fx248Subfx248(py, by);
  fixed248_t cpx = fx248Subfx248(px, cx);
  fixed248_t cpy = fx248Subfx248(py, cy);

  fixed248_t cross1 = fx248Subfx248(
    fx248Mulfx248(abx, apy),
    fx248Mulfx248(aby, apx)
  );
  fixed248_t cross2 = fx248Subfx248(
    fx248Mulfx248(bcx, bpy),
    fx248Mulfx248(bcy, bpx)
  );
  fixed248_t cross3 = fx248Subfx248(
    fx248Mulfx248(cax, cpy),
    fx248Mulfx248(cay, cpx)
  );

  bool_t hasNeg = (
    (cross1 < FIXED248_ZERO) ||
    (cross2 < FIXED248_ZERO) ||
    (cross3 < FIXED248_ZERO)
  );
  
  bool_t hasPos = (
    (cross1 > FIXED248_ZERO) ||
    (cross2 > FIXED248_ZERO) ||
    (cross3 > FIXED248_ZERO)
  );

  return !(hasNeg && hasPos);
}

collisionresult_t physicsCheckCircleTriangle(
  fixed248_t circleX, fixed248_t circleY, fixed248_t circleR,
  fixed248_t triX0, fixed248_t triY0,
  fixed248_t triX1, fixed248_t triY1,
  fixed248_t triX2, fixed248_t triY2
) {
  collisionresult_t result = { .hit = false };
  fixed248_t vx[3] = { triX0, triX1, triX2 };
  fixed248_t vy[3] = { triY0, triY1, triY2 };

  fixed248_t closestX = FIXED248_ZERO;
  fixed248_t closestY = FIXED248_ZERO;
  fixed248_t normalX  = FIXED248_ZERO;
  fixed248_t normalY  = FIXED248_ZERO;
  fixed248_t minDistSq = FIXED248_MAX;

  for(uint8_t i = 0; i < 3; ++i) {
    uint8_t j = (i + 1) % 3;

    fixed248_t testX, testY;
    physicsClosestPointOnSegment(
      vx[i], vy[i], vx[j], vy[j],
      circleX, circleY, &testX, &testY
    );

    fixed248_t dx = fx248Subfx248(circleX, testX);
    fixed248_t dy = fx248Subfx248(circleY, testY);
    fixed248_t distSq = fx248Addfx248(
      fx248Mulfx248(dx, dx),
      fx248Mulfx248(dy, dy)
    );

    if(distSq < minDistSq) {
      minDistSq = distSq;
      closestX = testX;
      closestY = testY;
      normalX = dx;
      normalY = dy;
    }
  }

  fixed248_t dist = fx248Sqrt(minDistSq);
  fixed248_t invDist = (
    (dist != FIXED248_ZERO) ?
    fx248Divfx248(FIXED248_ONE, dist) :
    FIXED248_ONE
  );

  fixed248_t nx = fx248Mulfx248(-normalX, invDist);
  fixed248_t ny = fx248Mulfx248(-normalY, invDist);

  if(physicsIsPointInTriangle(
    circleX, circleY, vx[0], vy[0], vx[1], vy[1], vx[2], vy[2]
  )) {
    result.hit = true;
    result.depth = fx248Subfx248(circleR, dist);
    result.normalX = nx;
    result.normalY = ny;
    return result;
  }

  if(dist < circleR) {
    result.hit = true;
    result.depth = fx248Subfx248(circleR, dist);
    result.normalX = nx;
    result.normalY = ny;
  }

  return result;
}
