/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#pragma once
#include "util/fixed.h"

typedef struct {
  bool_t hit;
  fixed248_t normalX, normalY;
  fixed248_t depth;
} collisionresult_t;

/**
 * Check for collision between two circles.
 * 
 * @param circle0x X coordinate of the first circle's center.
 * @param circle0y Y coordinate of the first circle's center.
 * @param circle0r Radius of the first circle.
 * @param circle1x X coordinate of the second circle's center.
 * @param circle1y Y coordinate of the second circle's center.
 * @param circle1r Radius of the second circle.
 * @return A collisionresult_t structure containing collision information.
 */
collisionresult_t physicsCheckCircleCircle(
  fixed248_t circle0x, fixed248_t circle0y, fixed248_t circle0r,
  fixed248_t circle1x, fixed248_t circle1y, fixed248_t circle1r
);

/**
 * Check for collision between a circle and an axis-aligned bounding box (AABB).
 * 
 * @param circleX X coordinate of the circle's center.
 * @param circleY Y coordinate of the circle's center.
 * @param circleR Radius of the circle.
 * @param aabb X coordinate of the AABB's top-left corner.
 * @param aabbY Y coordinate of the AABB's top-left corner.
 * @param aabbWidth Width of the AABB.
 * @param aabbHeight Height of the AABB.
 * @return A collisionresult_t structure containing collision information.
 */
collisionresult_t physicsCheckCircleAABB(
  fixed248_t circleX, fixed248_t circleY, fixed248_t circleR,
  fixed248_t aabb, fixed248_t aabbY,
  fixed248_t aabbWidth, fixed248_t aabbHeight
);

/**
 * Calculate the closest point on a line segment to a point.
 * 
 * @param ax X coordinate of the first endpoint of the segment.
 * @param ay Y coordinate of the first endpoint of the segment.
 * @param bx X coordinate of the second endpoint of the segment.
 * @param by Y coordinate of the second endpoint of the segment.
 * @param px X coordinate of the point.
 * @param py Y coordinate of the point.
 * @param outX Pointer to store the X coordinate of the closest point.
 * @param outY Pointer to store the Y coordinate of the closest point.
 */
void physicsClosestPointOnSegment(
  fixed248_t ax, fixed248_t ay,
  fixed248_t bx, fixed248_t by,
  fixed248_t px, fixed248_t py,
  fixed248_t *outX, fixed248_t *outY
);

/**
 * Check if a point is inside a triangle defined by three vertices.
 * 
 * @param px X coordinate of the point.
 * @param py Y coordinate of the point.
 * @param x0 X coordinate of the first vertex of the triangle.
 * @param y0 Y coordinate of the first vertex of the triangle.
 * @param x1 X coordinate of the second vertex of the triangle.
 * @param y1 Y coordinate of the second vertex of the triangle.
 * @param x2 X coordinate of the third vertex of the triangle.
 * @param y2 Y coordinate of the third vertex of the triangle.
 * @return true if the point is inside the triangle, false otherwise.
 */
bool_t physicsIsPointInTriangle(
  fixed248_t px, fixed248_t py,
  fixed248_t x0, fixed248_t y0,
  fixed248_t x1, fixed248_t y1,
  fixed248_t x2, fixed248_t y2
);

/**
 * Check for collision between a circle and a triangle.
 * 
 * @param circleX X coordinate of the circle's center.
 * @param circleY Y coordinate of the circle's center.
 * @param circleR Radius of the circle.
 * @param triX0 X coordinate of the first vertex of the triangle.
 * @param triY0 Y coordinate of the first vertex of the triangle.
 * @param triX1 X coordinate of the second vertex of the triangle.
 * @param triY1 Y coordinate of the second vertex of the triangle.
 * @param triX2 X coordinate of the third vertex of the triangle.
 * @param triY2 Y coordinate of the third vertex of the triangle.
 * @return A collisionresult_t structure containing collision information.
 */
collisionresult_t physicsCheckCircleTriangle(
  fixed248_t circleX, fixed248_t circleY, fixed248_t circleR,
  fixed248_t triX0, fixed248_t triY0,
  fixed248_t triX1, fixed248_t triY1,
  fixed248_t triX2, fixed248_t triY2
);