/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "fixed.h"
#include "assert/assert.h"

fixed248_t fx248Fromi32(const int32_t b) {
  return (fixed248_t)b << FIXED248_FRACTION_BITS;
}

fixed248_t fx248Fromu32(const uint32_t b) {
  return (fixed248_t)((int32_t)b << FIXED248_FRACTION_BITS);
}

fixed248_t fx248Fromf32(const float_t b) {
  return (fixed248_t)(b * (1 << FIXED248_FRACTION_BITS));
}

fixed248_t fx248Fromu16(const uint16_t b) {
  return (fixed248_t)((int32_t)b << FIXED248_FRACTION_BITS);
}

fixed248_t fx248Fromu8(const uint8_t b) {
  return (fixed248_t)((int32_t)b << FIXED248_FRACTION_BITS);
}



int32_t fx248Toi32(const fixed248_t a) {
  return a >> FIXED248_FRACTION_BITS;
}

uint32_t fx248Tou32(const fixed248_t a) {
  return (uint32_t)(a >> FIXED248_FRACTION_BITS);
}

float_t fx248Tof32(const fixed248_t a) {
  return (float_t)a / (1 << FIXED248_FRACTION_BITS);
}

uint16_t fx248Tou16(const fixed248_t a) {
  return (uint16_t)(a >> FIXED248_FRACTION_BITS);
}

uint8_t fx248Tou8(const fixed248_t a) {
  return (uint8_t)(a >> FIXED248_FRACTION_BITS);
}



fixed248_t fx248Addfx248(const fixed248_t a, const fixed248_t b) {
  return a + b;
}

fixed248_t fx248Addi32(const fixed248_t a, const int32_t b) {
  return fx248Addfx248(a, fx248Fromi32(b));
}

fixed248_t fx248Addu32(const fixed248_t a, const uint32_t b) {
  return fx248Addfx248(a, fx248Fromu32(b));
}

fixed248_t fx248Addf32(const fixed248_t a, const float_t b) {
  return fx248Addfx248(a, fx248Fromf32(b));
}



fixed248_t fx248Subfx248(const fixed248_t a, const fixed248_t b) {
  return a - b;
}

fixed248_t fx248Subi32(const fixed248_t a, const int32_t b) {
  return fx248Subfx248(a, fx248Fromi32(b));
}

fixed248_t fx248Subu32(const fixed248_t a, const uint32_t b) {
  return fx248Subfx248(a, fx248Fromu32(b));
}

fixed248_t fx248Subf32(const fixed248_t a, const float_t b) {
  return fx248Subfx248(a, fx248Fromf32(b));
}



fixed248_t fx248Mulfx248(const fixed248_t a, const fixed248_t b) {
  return (fixed248_t)(((int64_t)a * (int64_t)b) >> FIXED248_FRACTION_BITS);
}

fixed248_t fx248Muli32(const fixed248_t a, const int32_t b) {
  return (fixed248_t)(((int64_t)a * (int64_t)b) >> FIXED248_FRACTION_BITS);
}

fixed248_t fx248Mulu32(const fixed248_t a, const uint32_t b) {
  return (fixed248_t)(
    ((int64_t)a * (int64_t)(int32_t)b
  ) >> FIXED248_FRACTION_BITS);
}

fixed248_t fx248Mulf32(const fixed248_t a, const float_t b) {
  return (fixed248_t)((
    (int64_t)a * (int64_t)(b * (1 << FIXED248_FRACTION_BITS))
  ) >> FIXED248_FRACTION_BITS);
}



fixed248_t fx248Divfx248(const fixed248_t a, const fixed248_t b) {
  assertFalse(b == 0, "Division by zero in fx248Divfx248");
  return (fixed248_t)(((int64_t)a << FIXED248_FRACTION_BITS) / (int64_t)b);
}

fixed248_t fx248Divi32(const fixed248_t a, const int32_t b) {
  assertFalse(b == 0, "Division by zero in fx248Divi32");
  return (fixed248_t)(((int64_t)a << FIXED248_FRACTION_BITS) / (int64_t)b);
}

fixed248_t fx248Divu32(const fixed248_t a, const uint32_t b) {
  assertFalse(b == 0, "Division by zero in fx248Divu32");
  return (fixed248_t)(
    ((int64_t)a << FIXED248_FRACTION_BITS
  ) / (int64_t)(int32_t)b);
}

fixed248_t fx248Divf32(const fixed248_t a, const float_t b) {
  assertFalse(b == 0, "Division by zero in fx248Divf32");
  return (fixed248_t)((
    (int64_t)a << FIXED248_FRACTION_BITS
  ) / (int64_t)(b * (1 << FIXED248_FRACTION_BITS)));
}



fixed248_t fx248Floor(const fixed248_t a) {
  return a & ~((1 << FIXED248_FRACTION_BITS) - 1);
}

fixed248_t fx248Ceil(const fixed248_t a) {
  if(a & ((1 << FIXED248_FRACTION_BITS) - 1)) {
    return (a & ~((1 << FIXED248_FRACTION_BITS) - 1)) + (1 << FIXED248_FRACTION_BITS);
  }
  return a;
}

fixed248_t fx248Round(const fixed248_t a) {
  if(a & ((1 << (FIXED248_FRACTION_BITS - 1)) - 1)) {
    return (a & ~((1 << FIXED248_FRACTION_BITS) - 1)) + (1 << FIXED248_FRACTION_BITS);
  }
  return a & ~((1 << FIXED248_FRACTION_BITS) - 1);
}



uint32_t fx248Flooru32(const fixed248_t a) {
  return (uint32_t)((a >> FIXED248_FRACTION_BITS) & 0xFFFFFFFF);
}

uint32_t fx248Ceilu32(const fixed248_t a) {
  return (uint32_t)(((a + ((1 << FIXED248_FRACTION_BITS) - 1)) >> FIXED248_FRACTION_BITS) & 0xFFFFFFFF);
}

uint32_t fx248Roundu32(const fixed248_t a) {
  return (uint32_t)(((a + (1 << (FIXED248_FRACTION_BITS - 1))) >> FIXED248_FRACTION_BITS) & 0xFFFFFFFF);
}

fixed248_t fx248Sqrt(const fixed248_t a) {
  if(a == 0) return 0;

  fixed248_t y = a > FIXED248(1, 0) ? a : FIXED248(1, 0);
  fixed248_t last = 0;
  int max_iter = 16;
  while(y != last && max_iter-- > 0) {
    last = y;
    int32_t div = (int32_t)(((int64_t)a << FIXED248_FRACTION_BITS) / y);
    y = (y + div) >> 1;
  }
  return y;
}



fixed248_t fx248Max(const fixed248_t a, const fixed248_t b) {
  return (a > b) ? a : b;
}

fixed248_t fx248Min(const fixed248_t a, const fixed248_t b) {
  return (a < b) ? a : b;
}

fixed248_t fx248Clamp(
  const fixed248_t a,
  const fixed248_t min,
  const fixed248_t max
) {
  return (a < min) ? min : (a > max) ? max : a;
}